{$F+,O+}
UNIT grgencga;

{PURPOSE: constants, types, global variables and procedures
for general graphics use}

{March 7th 1993: Adapted to SVGA graphic modes by J.E.Mebius}

{October 5th 2000: PE24 items being phased out; in particular the global
variables PGte and GTmode are being phased out because graphics and text
modes coexisting in the Olivetti PE24 computer cannot coexist in other PCs;
the variable PGgr (in KSTYPVAR Unit) represents the screen mode:
PGgr = True <=> graphics mode; PGgr = false <=> text mode.}

{April 4th 2001: Introduction of CRT screen size parameter 10 <=> 17-inch
displays <=> 280 x 210 mm}

{April 23rd 2001: Unit TBGIGTOO incorporated into this unit GRGENCGA}
{April 23rd 2001: Procedure StandardSVGAColors moved to unit GRAPHDUM}
{April 29th 2001: Graphic-mode items in KSTYPVAR moved to this unit GRGENCGA}
{April 29th 2001: Unit GRAPHDUM incorporated into this unit GRGENCGA}

{October 13th 2003: Brown/Ochre choice for VGA instead of brown only}

INTERFACE USES dos, crt, kstypvar, graph;

{Graphics items not belonging to Borland's Graphic Interface:}
{Remark: Global types, constants and variables}

{============ HARDWARE-RELATED TYPES ============}

type GMDType   = {Graphics Mode Descriptor:}
     {Layout of parameters of TurboPASCAL InitGraph procedure}
     record
      graphdriver,
      graphmode:   integer;
      driverpath:  stringMaxL;
      origgd:      integer; {To retain original value of graphdriver parameter}
                            {because of "graphdriver := InstallUserDriver (..)}
     end; {GMDType}

type RGBType = record R, G, B: byte; end;

type SVGAPaletteType = array[0..255] of RGBType;

type SVGASetType = {Palette, Color Group, Color Index}
     record
      SVGAPalette:    SVGAPaletteType;
      SVGAcolorgroup: integer; {Current color group:}
                               {0..15, denoting index segments 0 <=> 0..15,
                               ... 15 <=> 240..255}
      SVGAcolorindex: integer; {0..15; offset within current color group}
      SVGAPalindex:   byte;    {Is set to 16 * group + index whenever needed}
     end; {SVGASetType}

type VGATextPalette = array[0..16] of byte;
     {[0]..[15]: text colors; [16]: border color}

{============ NOT-HARDWARE-RELATED TYPES ============}

type WorldType = {Window in 2D world}
     record
      {Left upper corner:}  x1, y1,
      {Right lower corner:} x2, y2: real;
     end; {WorldType}

type WoWinType = {Window in 2D world and viewport onto this window is mapped}
     record
      wo:       WorldType;     {Part of 2D world to be mapped onto viewport}
      wovp,                    {Viewport for window frame}
      acvp:     ViewportType;  {Viewport for actual drawing:}
                {redundant, but maintained for faster processing}
      lmargin,  {Left}
      tmargin,  {Top}
      rmargin,  {Right}
      bmargin:  {Bottom}
                integer; {Margins for actual viewport inside frame}
      wfcolorg, wfcolor: integer;  {World window frame color}
      wtcolorg, wtcolor: integer;  {World window text color}
      DrawFr:      boolean; {Draw window/viewport frame}
      DrawWC:      boolean; {Draw coordinates of 2D world window corners}
      DrawVC:      boolean; {Draw pixel coordinates of viewport corners}
      OnceSwitchW,
      OnceSwitchV: boolean; {To let corner coordinates be wiped just once}
     end; {WoWinType}


{============ HARDWARE-RELATED CONSTANTS ============}

const PGAtoSVGARatio = 4;
{PGA has levels 0...255 for R, G, B intensities; SVGA has levels 0...63}

{============ NOT-HARDWARE-RELATED CONSTANTS ============}

{Color group names:}
const CGAColors = 0; Pastels = 1;
const Draw = true; Wipe = false; {Parameter values for drawing procedures}
const CurrFont = DefaultFont;    {Font currently used}


{============ HARDWARE-RELATED VARIABLES ============}

var PGgr: boolean; {Check T/F state with calls to InitGraph and CloseGraph}

{============ NOT-HARDWARE-RELATED VARIABLES ============}

{Character dimensions in graphic mode:}
var PGCharSize:   integer; {Character size as specified in TPascal software}
var PGCharWidth:  integer; {Width in pixel units}
var PGCharHeight: integer; {Height in pixel units}
var PGLineWidth:  integer; {Maxiaml number of characters per line}

{Graphic counterparts of cursor coordinates and background color:}
var gcursor:    integer; {Number of lines; 0 to attain hidden graphic cursor}
var PGx,  PGy:  integer; {Cursor coordinates in character units,
                         like in text mode}
var PGpx, PGpy: integer; {Cursor coordinates in pixel units}
var Gbgcolor:   integer; {Background color in graphics mode}
var PGcolors:   boolean; {For controlling color/white drawing by ABPRDRAW}

{Preset line widths for graphics drawings:}
var PGELineWidth: word;  {Edges (in AB programs); lines in general}
var PGFLineWidth: word;  {Face edges (in AB programs)}


{============ HARDWARE-RELATED PROCEDURES AND FUNCTIONS ============}

procedure SetGMD (var GMD: GMDType; driver, mode: integer; path: StringMinL);
{---------------------------------------------------------------------------}
{PURPOSE: Prepare for InitGraph}

procedure EnterGraphicsMode ({io} var GMD: GMDType);
{--------------------------------------------------}
{PURPOSE: Initializes the graphics system according to the data in "GMD"}

procedure ReadSVGAColors
          (Start, Count: integer; {o} var Pal: SVGAPaletteType);
{--------------------------------------------------------------}
{PURPOSE: Reads consecutive entries from the hardware palette
into corresponding entries of Pal}

procedure SetSVGAColors
          (Start, Count: integer; {i} var Pal: SVGAPaletteType);
{--------------------------------------------------------------}
{PURPOSE: Writes consecutive entries from Pal
into corresponding entries of the hardware palette}

procedure SaveSVGAPalette ({o} var Pal: SVGAPaletteType);
{-------------------------------------------------------}
{PURPOSE: Copies all of the hardware palette to Pal}

procedure RestoreSVGAPalette ({i} var Pal: SVGAPaletteType);
{----------------------------------------------------------}
{PURPOSE: Copies Pal in its entirety to the hardware palette}


{============ NOT-HARDWARE-RELATED PROCEDURES AND FUNCTIONS ============}
{============ SVGA PROCEDURES AND FUNCTIONS ============}

function  GetColorGroup ({i} var SVGASet: SVGASetType): integer;
{--------------------------------------------------------------}
{Remark: from variable SVGASet, not directly from the hardware palette}

procedure SetColorGroup
          ({o} var SVGASet: SVGASetType; {i} var colorgroup: integer);
{--------------------------------------------------------------------}
{Remark: modifies glabal variable, does not affect the hardware palette}
{Perform RestoreSVGAPalette to let it take effect}

procedure GSetColor ({o} var SVGASet: SVGASetType; color: word);
{--------------------------------------------------------------}
{PURPOSE: Analogous to SetColor, but within color group instead of
entire palette; does not affect the hardware palette}
{Perform RestoreSVGAPalette to let it take effect}

procedure SetPalEntry
          ({o} var SVGAPalette: SVGAPaletteType; i, r, g, b: integer);
{--------------------------------------------------------------------}
{Set a single palette entry to specified red, green and blue levels}
{Perform RestoreSVGAPalette to let it take effect}

procedure ColorInterpolation
          ({o} var p: SVGAPaletteType;
          pe1, r1, g1, b1, pe2, r2, g2, b2: integer);
{---------------------------------------------------}
{PURPOSE: Fill entries number pe1 up to and including number pe2
of palette p with colors (r1, g1, b1) for pe1 ... (r2, g2, b2) for pe2
with linear interpolation for the entries in-between;
does not yet set palette}
{Perform RestoreSVGAPalette to let it take effect}
{PRE: pe1 < pe2}
{PENALTY: Division by zero}
{Remark: RGB intensities on the scale [0 ... 255]}

procedure StandardSVGAColors (var SVGAPalette: SVGAPaletteType);
{--------------------------------------------------------------}
{PURPOSE:}
{Groups 0 thru 3 and 6 thru 12 pertain to VGA colors}
{Group  0: 1/1 Saturated VGA colors:  0..15}
{Group  1: 1/2 Saturated VGA colors: 16..31}
{Group  2: 1/3 Saturated VGA colors: 32..47}
{Group  3: 1/4 Saturated VGA colors: 48..63}
{Group  4: Set  64.. 79 to 16 intensities of ochre:}
{Group  5: Set  80.. 95 to 16 intensities of mint-turquoise:}
{Group  6: Set  96..111 to intensities 0 (17) 255 of blue:}
{Group  7: Set 112..127 to intensities 0 (17) 255 of green:}
{Group  8: Set 128..143 to intensities 0 (17) 255 of cyan:}
{Group  9: Set 144..159 to intensities 0 (17) 255 of red:}
{Group 10: Set 160..175 to intensities 0 (17) 255 of magenta:}
{Group 11: Set 176..191 to intensities 0 (17) 255 of yellow:}
{Group 12: Set 192..207 to intensities 0 (17) 255 of gray/white:}
{Group 13: Set 208..223 to intensities of siena:}
{Group 14: Set 224..239 to intensities of Cherenkov-blue:}
{Group 15: Set 240..255 to intensities reddish gold:}

procedure SetStandardSVGASet ({o} var SVGASet: SVGASetType);
{----------------------------------------------------------}
{PURPOSE: Set standard SVGA palette, color group CGAColors and color white in
SVGASet record, but do not yet set the hardware palette}

procedure SetupGraphics
({i} var GMD: GMDType; {o} var SVGASet: SVGASetType;
{o} var vp: ViewportType; {i} StColors, StViewport: boolean);
{-----------------------------------------------------------}
{PURPOSE: Initializes the graphics system according to the data in "GMD";
sets color group to "cgacol", line style (global data in units);
sets SVGApalette to standard colors if desired;
sets viewport "vp" to INITGRAPH value if desired.}
{Adapted to SVGA graphic modes March 7th 1993 by J.E.Mebius}

procedure SetupGraphicsEx
({i} var GMD: GMDType; {o} var SVGASet: SVGASetType;
{o} var vp: ViewportType; {i} StColors, STViewport: boolean);
{-----------------------------------------------------------}
{PURPOSE: Like "SetupGraphics", but remembers color group.}

procedure SwitchColorGroup ({o} var SVGASet: SVGASetType);
{--------------------------------------------------------}
{PURPOSE: To switch between saturated and pastel CGA colors}


{============ WINDOW/VIEWPORT PROCEDURES AND FUNCTIONS ============}

procedure DefineWorld (var wo: WorldType; wox1, woy1, wox2, woy2: real);
{----------------------------------------------------------------------}
{PURPOSE: Defines window in 2D world to be mapped onto a viewport}

procedure DefineViewport
({o} var vp: ViewportType; vpx1, vpy1, vpx2, vpy2: integer; clipping: boolean);
{-----------------------------------------------------------------------------}
{PURPOSE: Defines a viewport for later use.}

procedure SetWindowWorld
          ({o} var WoWin: WoWinType; {i} var world: WorldType);
{-------------------------------------------------------------}
{PURPOSE: Binds 2D world parameter "world" to "WoWin"}

procedure SetWindowViewport
          ({o} var WoWin: WoWinType; {i} var vp: ViewportType);
{-------------------------------------------------------------}
{PURPOSE: Binds viewport parameter "vp" to "WoWin"}

procedure GetWindowViewport
          ({o} var vp: ViewportType; {i} var WoWin: WoWinType);
{-------------------------------------------------------------}
{PURPOSE: Retrieves viewport parameter "vp" from "WoWin"}

procedure SetWindowColors
          ({o} var WoWin: WoWinType;
          fcolorg, fcolor, tcolorg, tcolor: integer);
{---------------------------------------------------}
{PURPOSE: Sets colors of window frame and corner coordinates}

procedure SetWindowFrameControl ({o} var WoWin: WoWinType; DrwFr: boolean);
{--------------------------------------------------------------------------}
{PURPOSE: Controls drawing of window frame}

procedure SetWindowWoCoordControl ({o} var WoWin: WoWinType; DrwWC: boolean);
{---------------------------------------------------------------------------}
{PURPOSE: Controls writing of 2D world corner coordinates}

procedure SetWindowVpCoordControl ({o} var WoWin: WoWinType; DrwVC: boolean);
{---------------------------------------------------------------------------}
{PURPOSE: Controls writing of viewport corner coordinates}

procedure SetWindowControl
          ({o} var WoWin: WoWinType; DrwFr, DrwWC, DrwVC: boolean);
{------------------------------------------------------------------}
{PURPOSE: Controls drawing of window frame and writing of corner coordinates}

procedure SetWindowViewportMargin ({o} var WoWin: WoWinType; margin: integer);
{----------------------------------------------------------------------------}
{PURPOSE: Sets margin between window frame and actual viewport to be set
inside window}

procedure SetWindow
          ({o} var WoWin: WoWinType; {i} var world: WorldType;
          {i} var viewport: ViewportType;
          fcolorg, fcolor, tcolorg, tcolor: integer;
          DrwFr, DrwWC, DrwVC: boolean; margin: integer);
{------------------------------------------------------------}
{PURPOSE: Defines window and viewport for the 2D world.}

procedure DrawWindow
({i} var WoWin: WoWinType; {i} var SVGASet: SVGASetType; DrawWipe: boolean);
{--------------------------------------------------------------------------}
{PURPOSE: Draws or erases (draws in background color) frame of window in
2D world and coordinates of its corners; leaves with the actual viewport
just inside the frame just drawn}
{Remark: Protected by means of global variable "PGgr"}

procedure PGgron
(var GMD: GMDType; var SVGASet: SVGASetType; var WoWin: WoWinType);
{-----------------------------------------------------------------}
{PURPOSE: Forces graphics mode and redraws window}
{Remark: Protected by means of global variable "PGgr"}

procedure PGgroff;
{----------------}
{PURPOSE: Forces text mode}
{Remark: Protected by means of global variable "PGgr"}

procedure PGgronoff
(var GMD: GMDType; var SVGASet: SVGASetType; var WoWin: WoWinType);
{-----------------------------------------------------------------}
{PURPOSE: Switches between text and graphics modes}
{Remark: Protected by means of global variable "PGgr"}

procedure WindowFrOn
          ({o} var WoWin: WoWinType; {i} var SVGASet: SVGASetType);
{-----------------------------------------------------------------}
{PURPOSE: Turn drawing of frame ON}

procedure WindowFrOff
          ({o} var WoWin: WoWinType; {i} var SVGASet: SVGASetType);
{-----------------------------------------------------------------}
{PURPOSE: Turn drawing of frame OFF}

procedure WindowFrOnOff
          ({o} var WoWin: WoWinType; {i} var SVGASet: SVGASetType);
{-----------------------------------------------------------------}
{PURPOSE: Switch  drawing of frame}

procedure WindowWoCoOn
          ({o} var WoWin: WoWinType; {i} var SVGASet: SVGASetType);
{-----------------------------------------------------------------}
{PURPOSE: Turn writing 2D world corner coordinates ON}

procedure WindowWoCoOff
          ({o} var WoWin: WoWinType; {i} var SVGASet: SVGASetType);
{-----------------------------------------------------------------}
{PURPOSE: Turn writing 2D world corner coordinates OFF}

procedure WindowWoCoOnOff
          ({o} var WoWin: WoWinType; {i} var SVGASet: SVGASetType);
{-----------------------------------------------------------------}
{PURPOSE: Switch writing 2D world corner coordinates}

procedure WindowVpCoOn
          ({o} var WoWin: WoWinType; {i} var SVGASet: SVGASetType);
{-----------------------------------------------------------------}
{PURPOSE: Turn writing viewport corner coordinates ON}

procedure WindowVpCoOff
          ({o} var WoWin: WoWinType; {i} var SVGASet: SVGASetType);
{-----------------------------------------------------------------}
{PURPOSE: Turn writing viewport corner coordinates OFF}

procedure WindowVpCoOnOff
          ({o} var WoWin: WoWinType; {i} var SVGASet: SVGASetType);
{-----------------------------------------------------------------}
{PURPOSE: Switch writing viewport corner coordinates}

procedure StandardWorldSetting (var wo: WorldType; setting: integer);
{-------------------------------------------------------------------}
{PURPOSE: Sets window in 2D world to CRT size in pixel units (0...4)
or to physical CRT size (5, 9, 10; reductions: 6, 7, 8}
{setting =  0 ... 4: (SVGA mode parameter) length unit = pixel size}
{setting =  5: nearly the 15-inch CRT dimensions in cm}
{setting =  6, 7, 8:  3/4, 3/5, 3/6 reduction of setting = 5}
{setting =  9: mirror image of setting = 5}
{setting = 10: nearly the 17-inch CRT dimensions in cm}


IMPLEMENTATION


{============ HARDWARE-RELATED PROCEDURES AND FUNCTIONS ============}

procedure SetGMD (var GMD: GMDType; driver, mode: integer; path: StringMinL);
{---------------------------------------------------------------------------}
{PURPOSE: Prepare for InitGraph}
begin
 with GMD do
 begin
  graphdriver := driver; graphmode := mode; driverpath := path;
 end; {with GMD}
end; {SetGMD}

procedure EnterGraphicsMode ({io} var GMD: GMDType);
{--------------------------------------------------}
{PURPOSE: Initializes the graphics system according to the data in "GMD"}
begin
 {GMD.graphdriver = 1 ... 10 occupied by TurboPASCAL 6.0 settings;}
 {GMD.graphdriver = 11, 12, 13 occupied by SVGA}
 with GMD do
 begin
  {Retain original value of graphdriver parameter:} origgd := graphdriver;
  {If required the install SVGA driver:}
  case origgd of
  11: graphdriver := InstallUserDriver ('SVGA256', NIL);
  12: graphdriver := InstallUserDriver ('SVGA256M', NIL);
  13: graphdriver := InstallUserDriver ('SVGA32K', NIL);
  end; {case origgd}
  {Enter graphics mode and set accompanying variable PGgr:}
  initgraph (graphdriver, graphmode, driverpath); PGgr := true;
 end; {with GMD}
end; {EnterGraphicsMode}

procedure ReadSVGAColors
          (Start, Count: integer; {o} var Pal: SVGAPaletteType);
{--------------------------------------------------------------}
{PURPOSE: Reads consecutive entries from the hardware palette
into corresponding entries of Pal}
var Regs: Registers;
begin
 if PGgr
 then
 begin
  with Regs do
  begin
   AH := $10;
   AL := $17;
   BX := Start;
   CX := Count;
   ES := Seg (Pal);
   DX := Ofs (Pal);
  end;
  Intr ($10, Regs);
 end; {if PGgr}
end; {ReadSVGAColors}

procedure SetSVGAColors
          (Start, Count: integer; {i} var Pal: SVGAPaletteType);
{--------------------------------------------------------------}
{PURPOSE: Writes consecutive entries from Pal
into corresponding entries of the hardware palette}
var Regs: Registers;
begin
 if PGgr
 then
 begin
  with Regs do
  begin
   AH := $10;
   AL := $12;
   BX := Start;
   CX := Count;
   ES := Seg (Pal);
   DX := Ofs (Pal);
  end;
  Intr ($10, Regs);
 end; {if PGgr}
end; {SetSVGAColors}

procedure SaveSVGAPalette ({o} var Pal: SVGAPaletteType);
{-------------------------------------------------------}
{PURPOSE: Copies all of the hardware palette to Pal}
begin ReadSVGAColors (0, 256, Pal); end;

procedure RestoreSVGAPalette ({i} var Pal: SVGAPaletteType);
{----------------------------------------------------------}
{PURPOSE: Copies Pal in its entirety to the hardware palette}
begin SetSVGAColors (0, 256, Pal); end;

{============ NOT-HARDWARE-RELATED PROCEDURES AND FUNCTIONS ============}
{============ SVGA PROCEDURES AND FUNCTIONS ============}

function  GetColorGroup ({i} var SVGASet: SVGASetType): integer;
{--------------------------------------------------------------}
{Remark: from variable SVGASet, not directly from the hardware palette}
begin GetColorGroup := SVGASet.SVGAcolorgroup; end;

procedure SetColorGroup
          ({o} var SVGASet: SVGASetType; {i} var colorgroup: integer);
{--------------------------------------------------------------------}
{Remark: modifies glabal variable, does not affect the hardware palette}
{Perform RestoreSVGAPalette to let it take effect}
begin SVGASet.SVGAcolorgroup := colorgroup; end;

procedure GSetColor ({o} var SVGASet: SVGASetType; color: word);
{--------------------------------------------------------------}
{PURPOSE: Analogous to SetColor, but within color group instead of
entire palette; does not affect the hardware palette}
{Perform RestoreSVGAPalette to let it take effect}
begin
 if PGgr
 then with SVGASet do
 begin
  SVGAPalindex := 16 * SVGAcolorgroup + color;
  setcolor (SVGAPalindex);
 end; {if PRgr}
end; {GSetColor}

procedure SetPalEntry
          ({o} var SVGAPalette: SVGAPaletteType; i, r, g, b: integer);
{--------------------------------------------------------------------}
{Set a single palette entry to specified red, green and blue levels}
{Perform RestoreSVGAPalette to let it take effect}
var rr, gg, bb: byte;
begin
 rr := r div PGAtoSVGARatio;
 gg := g div PGAtoSVGARatio;
 bb := b div PGAtoSVGARatio;
 with SVGAPalette[i] do begin R := rr; G := gg; B := bb; end;
end; {SetPalEntry}

procedure ColorInterpolation
          ({o} var p: SVGAPaletteType;
          pe1, r1, g1, b1, pe2, r2, g2, b2: integer);
{---------------------------------------------------}
{PURPOSE: Fill entries number pe1 up to and including number pe2
of palette p with colors (r1, g1, b1) for pe1 ... (r2, g2, b2) for pe2
with linear interpolation for the entries in-between;
does not yet set palette}
{Perform RestoreSVGAPalette to let it take effect}
{PRE: pe1 < pe2}
{PENALTY: Division by zero}
{Remark: RGB intensities on the scale [0 ... 255]}
var i, r, g, b, n: integer;
begin
 n := pe2 - pe1;
 for i := 0 to n do
 begin
  SetPalEntry (p, pe1 + i,
  round (r1 + i * (r2 - r1) / n),
  round (g1 + i * (g2 - g1) / n),
  round (b1 + i * (b2 - b1) / n));
 end;
end; {ColorInterpolation}

procedure StandardSVGAColors (var SVGAPalette: SVGAPaletteType);
{--------------------------------------------------------------}
{PURPOSE:}
{Groups 0 thru 3 and 6 thru 12 pertain to VGA colors}
{Group  0: 1/1 Saturated VGA colors:  0..15}
{Group  1: 1/2 Saturated VGA colors: 16..31}
{Group  2: 1/3 Saturated VGA colors: 32..47}
{Group  3: 1/4 Saturated VGA colors: 48..63}
{Group  4: Set  64.. 79 to 16 intensities of ochre:}
{Group  5: Set  80.. 95 to 16 intensities of mint-turquoise:}
{Group  6: Set  96..111 to intensities 0 (17) 255 of blue:}
{Group  7: Set 112..127 to intensities 0 (17) 255 of green:}
{Group  8: Set 128..143 to intensities 0 (17) 255 of cyan:}
{Group  9: Set 144..159 to intensities 0 (17) 255 of red:}
{Group 10: Set 160..175 to intensities 0 (17) 255 of magenta:}
{Group 11: Set 176..191 to intensities 0 (17) 255 of yellow:}
{Group 12: Set 192..207 to intensities 0 (17) 255 of gray/white:}
{Group 13: Set 208..223 to intensities of siena:}
{Group 14: Set 224..239 to intensities of Cherenkov-blue:}
{Group 15: Set 240..255 to intensities reddish gold:}

{Intensity settings for partly saturated colors:}
{                            Low-Intensity  High-Intensity}
{(  0.. 15):  1/1 saturated:     0 and 144       0 and 255}
{( 16.. 31):  1/2 saturated:    72 and 144     128 and 255}
{( 32.. 47):  1/3 saturated:    96 and 144     170 and 255}
{( 48.. 63):  1/4 saturated:   108 and 144     192 and 255}
{Dark gray is realized by (80, 80, 80)}

var k, kk: integer;
begin
 {Assign CGA's saturated black..white, my pastel black..white to each of
 the segments (0..15) ... (48..63):}

 {Group 0: 1/1 Saturated colors: 0..15}
 kk := 0;
 {Low-intensity section}
      SetPalEntry (SVGAPalette, kk +  0,   0,   0,   0);
      SetPalEntry (SVGAPalette, kk +  1,   0,   0, 144);
      SetPalEntry (SVGAPalette, kk +  2,   0, 144,   0);
      SetPalEntry (SVGAPalette, kk +  3,   0, 144, 144);
      SetPalEntry (SVGAPalette, kk +  4, 144,   0,   0);
      SetPalEntry (SVGAPalette, kk +  5, 144,   0, 144);
      SetPalEntry (SVGAPalette, kk +  6, 144, 144,   0);
      SetPalEntry (SVGAPalette, kk +  7, 144, 144, 144);
 {High-intensity section}
      SetPalEntry (SVGAPalette, kk +  8,  80,  80,  80);
      SetPalEntry (SVGAPalette, kk +  9,   0,   0, 255);
      SetPalEntry (SVGAPalette, kk + 10,   0, 255,   0);
      SetPalEntry (SVGAPalette, kk + 11,   0, 255, 255);
      SetPalEntry (SVGAPalette, kk + 12, 255,   0,   0);
      SetPalEntry (SVGAPalette, kk + 13, 255,   0, 255);
      SetPalEntry (SVGAPalette, kk + 14, 255, 255,   0);
      SetPalEntry (SVGAPalette, kk + 15, 255, 255, 255);

 {Group 1: 1/2 Saturated colors: 16..31}
 kk := 16;
 {Low-intensity section}
      SetPalEntry (SVGAPalette, kk +  0,   0,   0,   0);
      SetPalEntry (SVGAPalette, kk +  1,  72,  72, 144);
      SetPalEntry (SVGAPalette, kk +  2,  72, 144,  72);
      SetPalEntry (SVGAPalette, kk +  3,  72, 144, 144);
      SetPalEntry (SVGAPalette, kk +  4, 144,  72,  72);
      SetPalEntry (SVGAPalette, kk +  5, 144,  72, 144);
      SetPalEntry (SVGAPalette, kk +  6, 144, 144,  72);
      SetPalEntry (SVGAPalette, kk +  7, 144, 144, 144);
 {High-intensity section}
      SetPalEntry (SVGAPalette, kk +  8,  80,  80,  80);
      SetPalEntry (SVGAPalette, kk +  9, 128, 128, 255);
      SetPalEntry (SVGAPalette, kk + 10, 128, 255, 128);
      SetPalEntry (SVGAPalette, kk + 11, 128, 255, 255);
      SetPalEntry (SVGAPalette, kk + 12, 255, 128, 128);
      SetPalEntry (SVGAPalette, kk + 13, 255, 128, 255);
      SetPalEntry (SVGAPalette, kk + 14, 255, 255, 128);
      SetPalEntry (SVGAPalette, kk + 15, 255, 255, 255);

 {Group 2: 1/3 Saturated colors: 32..47}
 kk := 32;
 {Low-intensity section}
      SetPalEntry (SVGAPalette, kk +  0,   0,   0,   0);
      SetPalEntry (SVGAPalette, kk +  1,  96,  96, 144);
      SetPalEntry (SVGAPalette, kk +  2,  96, 144,  96);
      SetPalEntry (SVGAPalette, kk +  3,  96, 144, 144);
      SetPalEntry (SVGAPalette, kk +  4, 144,  96,  96);
      SetPalEntry (SVGAPalette, kk +  5, 144,  96, 144);
      SetPalEntry (SVGAPalette, kk +  6, 144, 144,  96);
      SetPalEntry (SVGAPalette, kk +  7, 144, 144, 144);
 {High-intensity section}
      SetPalEntry (SVGAPalette, kk +  8,  80,  80,  80);
      SetPalEntry (SVGAPalette, kk +  9, 170, 170, 255);
      SetPalEntry (SVGAPalette, kk + 10, 170, 255, 170);
      SetPalEntry (SVGAPalette, kk + 11, 170, 255, 255);
      SetPalEntry (SVGAPalette, kk + 12, 255, 170, 170);
      SetPalEntry (SVGAPalette, kk + 13, 255, 170, 255);
      SetPalEntry (SVGAPalette, kk + 14, 255, 255, 170);
      SetPalEntry (SVGAPalette, kk + 15, 255, 255, 255);

 {Group 3: 1/4 Saturated colors: 48..63}
 kk := 48;
 {Low-intensity section}
      SetPalEntry (SVGAPalette, kk +  0,   0,   0,   0);
      SetPalEntry (SVGAPalette, kk +  1, 108, 108, 144);
      SetPalEntry (SVGAPalette, kk +  2, 108, 144, 108);
      SetPalEntry (SVGAPalette, kk +  3, 108, 144, 144);
      SetPalEntry (SVGAPalette, kk +  4, 144, 108, 108);
      SetPalEntry (SVGAPalette, kk +  5, 144, 108, 144);
      SetPalEntry (SVGAPalette, kk +  6, 144, 144, 108);
      SetPalEntry (SVGAPalette, kk +  7, 144, 144, 144);
 {High-intensity section}
      SetPalEntry (SVGAPalette, kk +  8,  80,  80,  80);
      SetPalEntry (SVGAPalette, kk +  9, 192, 192, 255);
      SetPalEntry (SVGAPalette, kk + 10, 192, 255, 192);
      SetPalEntry (SVGAPalette, kk + 11, 192, 255, 255);
      SetPalEntry (SVGAPalette, kk + 12, 255, 192, 192);
      SetPalEntry (SVGAPalette, kk + 13, 255, 192, 255);
      SetPalEntry (SVGAPalette, kk + 14, 255, 255, 192);
      SetPalEntry (SVGAPalette, kk + 15, 255, 255, 255);

 {Group 4: Set 64..79 to 16 intensities of yellow copper:}
 for k := 0 to 15 do SetPalEntry (SVGAPalette, 64 + k, 17*k, 16*k, 4*k);
 {Group 5: Set 80..95 to 16 intensities of mint-turquoise:}
 for k := 0 to 15 do SetPalEntry (SVGAPalette, 80 + k, 0, 17*k, 13*k);
 {Group 6: Set 96..111 to intensities 0 (17) 255 of blue:}
 for k := 0 to 15 do SetPalEntry (SVGAPalette, 96 + k, 0, 0, 17*k);
 {Group 7: Set 112..127 to intensities 0 (17) 255 of green:}
 for k := 0 to 15 do SetPalEntry (SVGAPalette, 112 + k, 0, 17*k, 0);
 {Group 8: Set 128..143 to intensities 0 (17) 255 of cyan:}
 for k := 0 to 15 do SetPalEntry (SVGAPalette, 128 + k, 0, 17*k, 17*k);
 {Group 9: Set 144..159 to intensities 0 (17) 255 of red:}
 for k := 0 to 15 do SetPalEntry (SVGAPalette, 144 + k, 17*k, 0, 0);
 {Group 10: Set 160..175 to intensities 0 (17) 255 of magenta:}
 for k := 0 to 15 do SetPalEntry (SVGAPalette, 160 + k, 17*k, 0, 17*k);
 {Group 11: Set 176..191 to intensities 0 (17) 255 of yellow:}
 for k := 0 to 15 do SetPalEntry (SVGAPalette, 176 + k, 17*k, 17*k, 0);
 {Group 12: Set 192..207 to intensities 0 (17) 255 of gray/white:}
 for k := 0 to 15 do SetPalEntry (SVGAPalette, 192 + k, 17*k, 17*k, 17*k);
 {Group 13: Set 208..223 to intensities of siena:}
 for k := 0 to 15 do SetPalEntry (SVGAPalette, 208 + k, 17*k, 9*k, 1*k);
 {Group 14: Set 224..239 to intensities of Cherenkov-blue:}
 for k := 0 to 15 do SetPalEntry (SVGAPalette, 224 + k, 8*k, 10*k, 17*k);
 {Group 15: Set 240..255 to intensities reddish gold:}
 for k := 0 to 15 do SetPalEntry (SVGAPalette, 240 + k, 17*k, 11*k, 1*k);
end; {StandardSVGAColors}

procedure SetStandardSVGASet ({o} var SVGASet: SVGASetType);
{----------------------------------------------------------}
{PURPOSE: Set standard SVGA palette, color group CGAColors and color white in
SVGASet record, but do not yet set the hardware palette}
var cg: integer;
begin
 with SVGASet do
 begin
  StandardSVGAColors (SVGAPalette);
  cg := CGAColors; SetColorGroup (SVGASet, cg);
  GSetColor (SVGASet, white);
 end; {with SVGASet}
end; {SetStandardSVGASet}

procedure SetupGraphics
({i} var GMD: GMDType; {o} var SVGASet: SVGASetType;
{o} var vp: ViewportType; {i} StColors, StViewport: boolean);
{-----------------------------------------------------------}
{PURPOSE: Initializes the graphics system according to the data in "GMD";
sets color group to "cgacol", line style (global data in units);
sets SVGApalette to standard colors if desired;
sets viewport "vp" to INITGRAPH value if desired.}
{Adapted to SVGA graphic modes March 7th 1993 by J.E.Mebius}
var cx, cy, dx, dy: integer;
var auxvp: ViewportType;
begin
 EnterGraphicsMode (GMD);
 {Specific SVGA processing:}
 if GMD.origgd > 10
 then
 begin
  if StColors then SetStandardSVGASet (SVGASet); {This includes color group
  being set to CGA and color being set to white}
  SetSVGAColors (0, 256, SVGASet.SVGAPalette);
 end; {GMD.origgd > 10}
 {Specific CGA...VGA processing:}
 if GMD.origgd <= 10 then SetColor (white);
 {Replace VGA brown (color 20 at index 6) by ochre (color 6):}
 if GetDriverName = 'EGAVGA' then SetPalette (6, 6);
 {General processing:}
 PGCharSize := 1;
 SetTextStyle (CurrFont, HorizDir, PGCharSize);
 PGCharWidth  :=  8 * PGCharSize;
 PGCharHeight := 10 * PGCharSize;
 GetViewSettings (auxvp);
 PGLineWidth := (auxvp.x2 + 1) div PGCharWidth;
 PGx := 1; PGy := 1;
 PGpx := PGCharWidth * (PGx - 1); PGpy := PGCharHeight * (PGy - 1);
 Gbgcolor := black;
 SetLineStyle (solidln, 0, normwidth);
 if StViewport then GetViewSettings (vp);
 (* Future feature: variable location and dimension of viewport
 cx := (vp.x1 + vp.x2) div 2; cy := (vp.y1 + vp.y2) div 2;
 dx := (vp.x1 - vp.x2) div 2; dy := (vp.y1 - vp.y2) div 2;
 dx := 2 * dx div 4; dy := 2 * dy div 4;
 with vp do
 begin
  x1 := cx + dx; y1 := cy + dy; x2 := cx - dx; y2 := cy - dy;
 end;
 *)
end; {SetupGraphics}

procedure SetupGraphicsEx
({i} var GMD: GMDType; {o} var SVGASet: SVGASetType;
{o} var vp: ViewportType; {i} StColors, STViewport: boolean);
{-----------------------------------------------------------}
{PURPOSE: Like "SetupGraphics", but remembers color group.}
var aux: integer;
begin
 aux := GetColorGroup (SVGASet);
 SetupGraphics (GMD, SVGASet, vp, StColors, STViewport);
 SetColorGroup (SVGASet, aux);
end; {SetupGraphicsEx}

procedure SwitchColorGroup ({o} var SVGASet: SVGASetType);
{--------------------------------------------------------}
{PURPOSE: To switch between saturated and pastel CGA colors}
var cg: integer;
begin
 case GetColorGroup (SVGASet) of
 CGAColors: begin  cg := Pastels; SetColorGroup (SVGASet, cg); end;
 Pastels: begin cg := CGAColors; SetColorGroup (SVGASet, cg); end;
 end; {case}
end; {SwitchColorGroup}


{============ WINDOW/VIEWPORT PROCEDURES AND FUNCTIONS ============}

procedure DefineWorld (var wo: WorldType; wox1, woy1, wox2, woy2: real);
{----------------------------------------------------------------------}
{PURPOSE: Defines window in 2D world to be mapped onto a viewport}
begin
 with wo do begin x1 := wox1; y1 := woy1; x2 := wox2; y2 := woy2; end;
end;

procedure DefineViewport
({o} var vp: ViewportType; vpx1, vpy1, vpx2, vpy2: integer; clipping: boolean);
{-----------------------------------------------------------------------------}
{PURPOSE: Defines a viewport for later use.}
begin
 with vp do
 begin x1 := vpx1; y1 := vpy1; x2 := vpx2; y2 := vpy2; clip := clipping; end;
end; {DefineViewport}

procedure SetWindowWorld
          ({o} var WoWin: WoWinType; {i} var world: WorldType);
{-------------------------------------------------------------}
{PURPOSE: Binds 2D world parameter "world" to "WoWin"}
begin
 WoWin.wo := world;
end; {SetWindowWorld}

procedure SetWindowViewport
          ({o} var WoWin: WoWinType; {i} var vp: ViewportType);
{-------------------------------------------------------------}
{PURPOSE: Binds viewport parameter "vp" to "WoWin"}
begin
 WoWin.wovp := vp;
end; {SetWindowViewport}

procedure GetWindowViewport
          ({o} var vp: ViewportType; {i} var WoWin: WoWinType);
{-------------------------------------------------------------}
{PURPOSE: Retrieves viewport parameter "vp" from "WoWin"}
begin
 vp := WoWin.wovp;
end; {GetWindowViewport}

procedure SetWindowColors
          ({o} var WoWin: WoWinType;
          fcolorg, fcolor, tcolorg, tcolor: integer);
{---------------------------------------------------}
{PURPOSE: Sets colors of window frame and corner coordinates}
begin
 with WoWin do
 begin
  {World window frame color:}
  wfcolorg := fcolorg;
  wfcolor  := fcolor;
  {World window text color:}
  wtcolorg := tcolorg;
  wtcolor  := tcolor;
 end; {with WoWin}
end; {SetWindowColors}

procedure SetWindowFrameControl ({o} var WoWin: WoWinType; DrwFr: boolean);
{-------------------------------------------------------------------------}
{PURPOSE: Controls drawing of window frame}
begin
 with WoWin do
 begin
  {Control variables:}
  DrawFr := DrwFr;   {draw window/viewport frame}
 end; {with WoWin}
end; {SetWindowFrameControl}

procedure SetWindowWoCoordControl ({o} var WoWin: WoWinType; DrwWC: boolean);
{---------------------------------------------------------------------------}
{PURPOSE: Controls writing of 2D world corner coordinates}
begin
 with WoWin do
 begin
  {Control variables:}
  DrawWC  := DrwWC;    {draw coordinates of 2D world window corners}
  OnceSwitchW := true; {if DrawWC became false then still wipe coordinates at first occasion}
 end; {with WoWin}
end; {SetWindowWoCoordControl}

procedure SetWindowVpCoordControl ({o} var WoWin: WoWinType; DrwVC: boolean);
{---------------------------------------------------------------------------}
{PURPOSE: Controls writing of viewport corner coordinates}
begin
 with WoWin do
 begin
  {Control variables:}
  DrawVC  := DrwVC;    {draw pixel coordinates of viewport corners}
  OnceSwitchV := true; {if DrawVC became false then still wipe coordinates at first occasion}
 end; {with WoWin}
end; {SetWindowVpCoordControl}

procedure SetWindowControl
          ({o} var WoWin: WoWinType; DrwFr, DrwWC, DrwVC: boolean);
{-----------------------------------------------------------------}
{PURPOSE: Controls drawing of window frame and writing of corner coordinates}
begin
 SetWindowFrameControl   (WoWin, DrwFr);
 SetWindowWoCoordControl (WoWin, DrwWC);
 SetWindowVpCoordControl (WoWin, DrwVC);
end; {SetWindowControl}

procedure SetWindowViewportMargin ({o} var WoWin: WoWinType; margin: integer);
{----------------------------------------------------------------------------}
{PURPOSE: Sets margin between window frame and actual viewport to be set
inside window}
begin
 with WoWin do
  begin
   {Margins for actual viewport inside frame:}
   lmargin := margin;
   tmargin := margin;
   rmargin := margin;
   bmargin := margin;
  end; {with WoWin}
end; {SetWindowViewportMargin}

procedure SetWindow
          ({o} var WoWin: WoWinType; {i} var world: WorldType;
          {i} var viewport: ViewportType;
          fcolorg, fcolor, tcolorg, tcolor: integer;
          DrwFr, DrwWC, DrwVC: boolean; margin: integer);
{------------------------------------------------------------}
{PURPOSE: Defines window and viewport for the 2D world.}
begin
 (*
 DefineWorld (world, wox1, woy1, wox2, woy2);
 DefineViewport (viewport, vpx1, vpy1, vpx2, vpy2, true);
 *)
 SetWindowWorld (WoWin, world);
 SetWindowViewport (WoWin, viewport);
 SetWindowColors (WoWin, fcolorg, fcolor, tcolorg, tcolor);
 SetWindowControl (WoWin, DrwFr, DrwWC, DrwVC);
 SetWindowViewportMargin (WoWin, margin);
end; {SetWindow}

procedure removeblanks (var s: stringMinL);
{-----------------------------------------}
{PRUPOSE: Removes all blanks from string s}
var h: stringMinL; i: integer;
begin
 h := '';
 for i := 1 to length (s) do if s[i] <> ' ' then h := h + s[i];
 s := h;
end;

procedure WriteWoCornerCoordinates
({i} var WoWin: WoWinType; {i} var SVGASet: SVGASetType;
wonum, dec: integer; colorc: integer; DrawWipe: boolean);
{-------------------------------------------------------}
var vwidth, vheight: integer; {Viewport}
var woxstr1, woystr1, woxstr2, woystr2: stringMinL;
var wolbstr, wolostr, worbstr, worostr: stringMedL;
var blankstring: stringMinL; var i: integer;
begin
 settextstyle (CurrFont, horizdir, PGCharSize);
 {Use string of solid blanks in background color for later wiping actions}
 blankstring[0] := chr (MinL); for i := 1 to MinL do blankstring[i] := #219;
 with WoWin do
 begin
  with wo do
  begin
   str (x1:wonum:dec, woxstr1); str (y1:wonum:dec, woystr1);
   str (x2:wonum:dec, woxstr2); str (y2:wonum:dec, woystr2);
  end; {with wo}
  removeblanks (woxstr1); removeblanks (woystr1);
  removeblanks (woxstr2); removeblanks (woystr2);
  wolbstr := '(' + woxstr1 + ', ' + woystr1 + ')';
  wolostr := '(' + woxstr1 + ', ' + woystr2 + ')';
  worbstr := '(' + woxstr2 + ', ' + woystr1 + ')';
  worostr := '(' + woxstr2 + ', ' + woystr2 + ')';
  with wovp do begin vwidth := x2 - x1; vheight := y2 - y1; end;
 end; {with WoWin}
 {Write coordinates:}
 GSetColor (SVGASet, colorc);
 {Left Top:}
 settextjustify (lefttext, toptext);
 if DrawWipe = Wipe
 then
 begin
  GSetColor (SVGASet, black);
  outtextxy (4, 4, blankstring);
 end; {Wipe}
 if DrawWipe = Draw
 then
 begin
  GSetColor (SVGASet, colorc);
  outtextxy (4, 4, wolbstr);
 end; {Draw}
 {Right Top:}
 settextjustify (righttext, toptext);
 if DrawWipe = Wipe
 then
 begin
  GSetColor (SVGASet, black);
  outtextxy (vwidth - 4, 4, blankstring);
 end; {Wipe}
 if DrawWipe = Draw
 then
 begin
  GSetColor (SVGASet, colorc);
  outtextxy (vwidth - 4, 4, worbstr);
 end; {Draw}
 {Left Bottom:}
 settextjustify (lefttext, bottomtext);
 if DrawWipe = Wipe
 then
 begin
  GSetColor (SVGASet, black);
  outtextxy (4, vheight - 4, blankstring);
 end; {Wipe}
 if DrawWipe = Draw
 then
 begin
  GSetColor (SVGASet, colorc);
  outtextxy (4, vheight - 4, wolostr);
 end; {Draw}
 {Right Bottom:}
 settextjustify (righttext, bottomtext);
 if DrawWipe = Wipe
 then
 begin
  GSetColor (SVGASet, black);
  outtextxy (vwidth - 4, vheight - 4, blankstring);
 end; {Wipe}
 if DrawWipe = Draw
 then
 begin
  GSetColor (SVGASet, colorc);
  outtextxy (vwidth - 4, vheight - 4, worostr);
 end; {Draw}
end; {WriteWoCornerCoordinates}

procedure WriteVpCornerCoordinates
({i} var WoWin: WoWinType; {i} var SVGASet: SVGASetType;
vpnum: integer; colorc: integer; DrawWipe: boolean);
{------------------------------------------------------}
var vwidth, vheight: integer; {Viewport}
var vpxstr1, vpystr1, vpxstr2, vpystr2: stringMinL;
var vplbstr, vplostr, vprbstr, vprostr: stringMedL;
var blankstring: stringMinL; var i: integer;
begin
 settextstyle (CurrFont, horizdir, PGCharSize);
 {Use string of solid blanks in background color for later wiping actions}
 blankstring[0] := chr (MinL); for i := 1 to MinL do blankstring[i] := #219;
 with WoWin do
 begin
  with wovp do
  begin
   str (x1:vpnum, vpxstr1); str (y1:vpnum, vpystr1);
   str (x2:vpnum, vpxstr2); str (y2:vpnum, vpystr2);
  end; {with wovp}
  removeblanks (vpxstr1); removeblanks (vpystr1);
  removeblanks (vpxstr2); removeblanks (vpystr2);
  vplbstr := '(' + vpxstr1 + ', ' + vpystr1 + ')';
  vplostr := '(' + vpxstr1 + ', ' + vpystr2 + ')';
  vprbstr := '(' + vpxstr2 + ', ' + vpystr1 + ')';
  vprostr := '(' + vpxstr2 + ', ' + vpystr2 + ')';
  with wovp do begin vwidth := x2 - x1; vheight := y2 - y1; end;
 end; {with WoWin}
 {Write coordinates:}
 GSetColor (SVGASet, colorc);
 {Left Top:}
 settextjustify (lefttext, toptext);
 if DrawWipe = Wipe
 then
 begin
  GSetColor (SVGASet, black);
  outtextxy (4, 4 + PGCharHeight, blankstring);
 end; {Wipe}
 if DrawWipe = Draw
 then
 begin
  GSetColor (SVGASet, colorc);
  outtextxy (4, 4 + PGCharHeight, vplbstr);
 end; {Draw}
 {Right Top:}
 settextjustify (righttext, toptext);
 if DrawWipe = Wipe
 then
 begin
  GSetColor (SVGASet, black);
  outtextxy (vwidth - 4, 4 + PGCharHeight, blankstring);
 end; {Wipe}
 if DrawWipe = Draw
 then
 begin
  GSetColor (SVGASet, colorc);
  outtextxy (vwidth - 4, 4 + PGCharHeight, vprbstr);
 end; {Draw}
 {Left Bottom:}
 settextjustify (lefttext, bottomtext);
 if DrawWipe = Wipe
 then
 begin
  GSetColor (SVGASet, black);
  outtextxy (4, vheight - 4 - PGCharHeight, blankstring);
 end; {Wipe}
 if DrawWipe = Draw
 then
 begin
  GSetColor (SVGASet, colorc);
  outtextxy (4, vheight - 4 - PGCharHeight, vplostr);
 end; {Draw}
 {Right Bottom:}
 settextjustify (righttext, bottomtext);
 if DrawWipe = Wipe
 then
 begin
  GSetColor (SVGASet, black);
  outtextxy (vwidth - 4, vheight - 4 - PGCharHeight, blankstring);
 end; {Wipe}
 if DrawWipe = Draw
 then
 begin
  GSetColor (SVGASet, colorc);
  outtextxy (vwidth - 4, vheight - 4 - PGCharHeight, vprostr);
 end; {Draw}
end; {WriteVpCornerCoordinates}

procedure DrawWindow
({i} var WoWin: WoWinType; {i} var SVGASet: SVGASetType; DrawWipe: boolean);
{--------------------------------------------------------------------------}
{PURPOSE: Draws or erases (draws in background color) frame of window on
2D world and coordinates of its corners; leaves with the actual viewport
just inside the frame just drawn}
{Remark: Protected by means of global variable "PGgr"}
var colorb: integer;
begin
 if PGgr
 then
 begin
  with WoWin do
  begin
   with wovp do SetViewport (x1, y1, x2, y2, ClipOn);
   case DrawWipe of
   Draw: if PGcolors then colorb := wfcolor else colorb := wfcolor or 7;
   Wipe: colorb := black;
   end; {case DrawWipe}
   SetColorGroup (SVGASet, wfcolorg);
   if DrawFr then GSetColor (SVGASet, colorb) else GSetColor (SVGASet, black);
   with wovp do rectangle (0, 0, x2 - x1, y2 - y1);
   case DrawWipe of
   Draw: if PGcolors then colorb := wtcolor else colorb := wtcolor or 7;
   Wipe: colorb := black;
   end; {case DrawWipe}
   SetColorGroup (SVGASet, wtcolorg);
   {Write or erase 2D world corner coordinates:}
   if DrawWC
   then
   begin
    WriteWoCornerCoordinates (WoWin, SVGASet, 8, 1, colorb, Draw);
    OnceSwitchW := true;
   end {if DrawWC}
   else
   begin
    {Wipe just once corner coordinates:}
    if OnceSwitchW
    then
    begin
     WriteWoCornerCoordinates (WoWin, SVGASet, 8, 1, colorb, Wipe);
     OnceSwitchW := false;
    end; {if OnceSwitchW}
   end; {if not DrawWC}
   {Write or erase world viewport corner coordinates:}
   if DrawVC
   then
   begin
    WriteVpCornerCoordinates (WoWin, SVGASet, 4, colorb, Draw);
    OnceSwitchV := true;
   end {if DrawVC}
   else
   begin
    {Wipe just once corner coordinates:}
    if OnceSwitchV
    then
    begin
     WriteVpCornerCoordinates (WoWin, SVGASet, 4, colorb, Wipe);
     OnceSwitchV := false;
    end; {if OnceSwitchV}
   end; {if not DrawVC}
   {Set actual viewport inside world viewport and store its parameters in "acvp":}
   with wovp do SetViewport (x1 + lmargin, y1 + tmargin, x2 - rmargin, y2 - bmargin, ClipOn);
   GetViewSettings (acvp);
  end; {with WoWin}
 end; {if PGgr}
end; {DrawWindow}

procedure PGgron
(var GMD: GMDType; var SVGASet: SVGASetType; var WoWin: WoWinType);
{-----------------------------------------------------------------}
{PURPOSE: Forces graphics mode and redraws window}
{Remark: Protected by means of global variable "PGgr"}
begin
 if not PGgr
 then
 begin
  PGgr := true;
  SetupGraphics (GMD, SVGASet, WoWin.wovp,
  {Do not modify SVGA palette and viewport:} false, false);
  DrawWindow (WoWin, SVGASet, Draw);
 end;
end;

procedure PGgroff;
{----------------}
{PURPOSE: Forces text mode}
{Remark: Protected by means of global variable "PGgr"}
begin
 if PGgr
 then
 begin
  {Replace VGA ochre (color 6 at index 6) by brown (color 20):}
  if GetDriverName = 'EGAVGA' then SetPalette (6, 20);
  PGgr := false;
  closegraph;
 end;
end;

procedure PGgronoff
(var GMD: GMDType; var SVGASet: SVGASetType; var WoWin: WoWinType);
{-----------------------------------------------------------------}
{PURPOSE: Switches between text and graphics modes}
{Remark: Protected by means of global variable "PGgr"}
begin
 case PGgr of
 TRUE:  PGgroff;
 FALSE: PGgron (GMD, SVGASet, WoWin);
 end;
end;

procedure WindowFrOn
          ({o} var WoWin: WoWinType; {i} var SVGASet: SVGASetType);
{-----------------------------------------------------------------}
{PURPOSE: Turn drawing of frame ON}
begin
 with WoWin do DrawFr := true;
 if PGgr then DrawWindow (WoWin, SVGASet, Draw);
end; {WindowFrOn}

procedure WindowFrOff
          ({o} var WoWin: WoWinType; {i} var SVGASet: SVGASetType);
{-----------------------------------------------------------------}
{PURPOSE: Turn drawing of frame OFF}
begin
 with WoWin do DrawFr := false;
 if PGgr then DrawWindow (WoWin, SVGASet, Draw);
end; {WindowFrOff}

procedure WindowFrOnOff
          ({o} var WoWin: WoWinType; {i} var SVGASet: SVGASetType);
{-----------------------------------------------------------------}
{PURPOSE: Switch  drawing of frame}
begin
 with WoWin do DrawFr := not DrawFr;
 if PGgr then DrawWindow (WoWin, SVGASet, Draw);
end; {WindowFrOnOff}

procedure WindowWoCoOn
          ({o} var WoWin: WoWinType; {i} var SVGASet: SVGASetType);
{-----------------------------------------------------------------}
{PURPOSE: Turn writing 2D world corner coordinates ON}
begin
 with WoWin do begin DrawWC := true; OnceSwitchW := true; end;
 if PGgr then DrawWindow (WoWin, SVGASet, Draw);
end; {WindowWoCoOn}

procedure WindowWoCoOff
          ({o} var WoWin: WoWinType; {i} var SVGASet: SVGASetType);
{-----------------------------------------------------------------}
{PURPOSE: Turn writing 2D world corner coordinates OFF}
begin
 with WoWin do begin DrawWC := false; OnceSwitchW := true; end;
 if PGgr then DrawWindow (WoWin, SVGASet, Draw);
end; {WindowWoCoOff}

procedure WindowWoCoOnOff
          ({o} var WoWin: WoWinType; {i} var SVGASet: SVGASetType);
{-----------------------------------------------------------------}
{PURPOSE: Switch writing 2D world corner coordinates}
begin
 with WoWin do begin DrawWC := not DrawWC; OnceSwitchW := true; end;
 if PGgr then DrawWindow (WoWin, SVGASet, Draw);
end; {WindowWoCoOnOff}

procedure WindowVpCoOn
          ({o} var WoWin: WoWinType; {i} var SVGASet: SVGASetType);
{-----------------------------------------------------------------}
{PURPOSE: Turn writing viewport corner coordinates ON}
begin
 with WoWin do DrawVC := true;
 if PGgr then DrawWindow (WoWin, SVGASet, Draw);
end; {WindowVpCoOn}

procedure WindowVpCoOff
          ({o} var WoWin: WoWinType; {i} var SVGASet: SVGASetType);
{-----------------------------------------------------------------}
{PURPOSE: Turn writing viewport corner coordinates OFF}
begin
 with WoWin do DrawVC := false;
 if PGgr then DrawWindow (WoWin, SVGASet, Draw);
end; {WindowVpCoOff}

procedure WindowVpCoOnOff
          ({o} var WoWin: WoWinType; {i} var SVGASet: SVGASetType);
{-----------------------------------------------------------------}
{PURPOSE: Switch writing viewport corner coordinates}
begin
 with WoWin do DrawVC := not DrawVC;
 if PGgr then DrawWindow (WoWin, SVGASet, Draw);
end; {WindowVpCoOnOff}

procedure StandardWorldSetting (var wo: WorldType; setting: integer);
{-------------------------------------------------------------------}
{PURPOSE: Sets window in 2D world to CRT size in pixel units (0...4)
or to physical CRT size (5, 9, 10; reductions: 6, 7, 8}
{setting =  0 ... 4: (SVGA mode parameter) length unit = pixel size}
{setting =  5: nearly the 15-inch CRT dimensions in cm}
{setting =  6, 7, 8:  3/4, 3/5, 3/6 reduction of setting = 5}
{setting =  9: mirror image of setting = 5}
{setting = 10: nearly the 17-inch CRT dimensions in cm}
begin
 with wo do
 case setting of
  0: begin
      x1 := -160.000; x2 := - x1 - 1; y2 := -100.000; y1 := - y2 - 1;
     end;
  1: begin
      x1 := -320.000; x2 := - x1 - 1; y2 := -200.000; y1 := - y2 - 1;
     end;
  2: begin
      x1 := -320.000; x2 := - x1 - 1; y2 := -240.000; y1 := - y2 - 1;
     end;
  3: begin
      x1 := -400.000; x2 := - x1 - 1; y2 := -300.000; y1 := - y2 - 1;
     end;
  4: begin
      x1 := -512.000; x2 := - x1 - 1; y2 := -384.000; y1 := - y2 - 1;
     end;
  5: begin
      x1 := -12.000; x2 := - x1; y1 := 9.000; y2 := - y1;
     end;
  6: begin
      x1 := -16.000; x2 := - x1; y1 := 12.000; y2 := - y1;
     end;
  7: begin
      x1 := -20.000; x2 := - x1; y1 := 15.000; y2 := - y1;
     end;
  8: begin
      x1 := -24.000; x2 := - x1; y1 := 18.000; y2 := - y1;
     end;
  9: begin
      x1 := 12.000; x2 := - x1; y1 := 9.000; y2 := - y1;
     end;
 10: begin
      x1 := -14.000; x2 := - x1; y1 := 10.500; y2 := - y1;
     end;
 end; {with wo ... case}
end; {StandardWorldSetting}

BEGIN
 (*write ('GRGENCGA v1.42-1995 ');*)
 (*write ('GRGENCGA v1.43-1999 ');*)
 (*write ('GRGENCGA v1.46-1999 ');*)
 (*write ('GRGENCGA v1.47-2000 ');*)
 (*write ('GRGENCGA v1.48-2000 ');*)
 (*write ('GRGENCGA v1.49-2001 ');*)
 PGgr := false;
 PGCharSize := 1;
 PGCharWidth  :=  8 * PGCharSize;
 PGCharHeight := 10 * PGCharSize;
 PGLineWidth := 80;
 gcursor := 2; {Graphic cursor initially bold}
 PGx := 1; PGy := 1;
 PGpx := PGCharWidth * (PGx - 1); PGpy := PGCharHeight * (PGy - 1);
 Gbgcolor := black;
 PGcolors := true;
 PGELineWidth := NormWidth;
 PGFLineWidth := NormWidth;
 write ('GRGENCGA v2.00-2001 ');
END.
